
#include "io.h"
#include "vfh.h"
#include "config.h"
#include "lib.h"


/***************************************
LocalMap
***************************************/


LocalMap::LocalMap(int w, int h, int size, float cellSize): w(w), h(h), size(size), cellSize(cellSize)
{
	map = new CertaintyCell*[w];
	for(int j = 0; j < w; j++) map[j] = new CertaintyCell[h]; // cells are initialized by the constructor
}

LocalMap::~LocalMap()
{
	for(int i = 0; i < w; i++) delete[] map[i];
	delete[] map;
}




void LocalMap::shift(int x, int y, bool memory)
{
	

}


/***************************************
PolarMap
***************************************/


PolarMap::PolarMap(int size): size(size)
{
	map = new float[size];
	tmp_map = new float[size];
}

PolarMap::~PolarMap()
{
	delete[] map;
	delete[] tmp_map;
}

void PolarMap::compute(LocalMap &lm, RobotState &rs)
{
	int activex = (int)(rs.x/lm.cellSize) - config.active_window_size/2;
	int activey = (int)(rs.y/lm.cellSize) - config.active_window_size/2;
	float dmax = sqrt(2)*config.active_window_size/2+1;
	
	// filling
	for(int i = 0; i < size; i++) tmp_map[i] = 0;
	for(int x = 0, xc = -config.active_window_size/2; x < config.active_window_size; x++, xc++)
	{
		if (x <= -activex || x+activex>=lm.w) continue;
		for(int y = 0, yc = -config.active_window_size/2; y < config.active_window_size; y++, yc++)
		{
			if (y <= -activey && y+activey>=lm.h) continue;
			float angle = atan2(yc,xc)*360/(2*M_PI); // TABLE !!
			if (angle < 0) angle += 360;
			int sector = (int)(angle*size/360);
			int val = lm.map[activex+x][activey+y].obst;
			float dist = sqrt(xc*xc + yc*yc);
			tmp_map[sector] += (val*val)*(dmax - dist);
		}
	}
	

	// prevent from getting out of the map
	if (rs.x < 40) for(int i = 1*size/4; i < 3*size/4; i++) map[i] = config.histo_thres*1.1;
	if (rs.y < 40) for(int i = 2*size/4; i < 4*size/4; i++) map[i] = config.histo_thres*1.1;
	if (rs.x > lm.w*lm.cellSize-40) 
	{ for(int i = 0*size/4; i < 1*size/4; i++) map[i] = config.histo_thres*1.1;
		for(int i = 3*size/4; i < 4*size/4; i++) map[i] = config.histo_thres*1.1; }
	if (rs.y > lm.h*lm.cellSize-40) for(int i = 0*size/4; i < 2*size/4; i++) map[i] = config.histo_thres*1.1;

	// smoothing
	for(int i = 0; i < size; i++) map[i] = 0;
	for(int i = 0; i < size; i++)
	{
		for(int j = 1; j <= config.smooth_polar_width; j++)
			map[i] += (config.smooth_polar_width-j+1)*(tmp_map[(i-j+size)%size] + tmp_map[(i+j)%size]);
		map[i] += (config.smooth_polar_width+1) * tmp_map[i];
		map[i] /= (config.smooth_polar_width+1)*(config.smooth_polar_width+1);
	}
	
}

float PolarMap::getDirection(RobotState &rs, float xGoal, float yGoal, float histo_thres)
{
	float goal_angle = atan2(yGoal-rs.y,xGoal-rs.x)*360/(2*M_PI);
	if (goal_angle < 0) goal_angle += 360;
	
	// find closest free sector
	int sector = (int)(goal_angle*size/360);
	int valley_plus, valley_minus;
	for(int i = 0; i < size/2; i++)
	{
		if (map[mod(sector+i,size)] < histo_thres) { valley_plus = valley_minus = mod(sector+i,size); break; }
		if (map[mod(sector-i,size)] < histo_thres) { valley_plus = valley_minus = mod(sector-i,size); break; }
	}
	
	// find boundaries of the valley the sector belongs to
	int valley_minus_m1 = mod(valley_minus-1,size);
	int valley_plus_p1  = mod(valley_plus+1,size);
	while ((map[valley_plus_p1] < histo_thres) && (valley_plus_p1 != valley_minus)) 
		{ valley_plus = valley_plus_p1; valley_plus_p1 = mod(valley_plus+1,size); }
	while ((map[valley_minus_m1] < histo_thres) && (valley_plus != valley_minus_m1)) 
		{ valley_minus = valley_minus_m1; valley_minus_m1 = mod(valley_minus-1,size); }
		
	// if no blocked angles, go to obstacle
	if (mod(valley_plus+1, size) == valley_minus) return sector*360/size;
	// if narrow valley, then middle
	if (mod(valley_plus-valley_minus, size)+1 < config.wide_valley_size)
		return avg_mod(valley_minus, valley_plus, size)*360/size;
	// if wide valley, goal outside, then near goal with margin
	if (map[sector] >= histo_thres)
		if (mod(valley_minus-sector, size) < mod(sector-valley_plus, size))
			return mod(valley_minus+config.wide_valley_size/2, size)*360/size; else
			return mod(valley_plus -config.wide_valley_size/2, size)*360/size;
	// if wide valley, goal inside, enough margin with goal, then goal
	if (mod(valley_plus-sector, size) > config.wide_valley_size/2 && mod(sector-valley_minus, size) > config.wide_valley_size/2)
		return sector*360/size;
	// if wide valley, goal inside, not enough margin with goal, then goal with margin
	if (mod(valley_plus-sector, size) > mod(sector-valley_minus, size))
		return mod(valley_minus+config.wide_valley_size/2, size)*360/size; else
		return mod(valley_plus -config.wide_valley_size/2, size)*360/size;
}

float PolarMap::getSpeed(RobotState &rs, float xGoal, float yGoal, bool stop_at_goal, float speed_reduction)
{
	int sector = (int)(rs.theta*size/360);
	float v = config.max_speed * (1 - min(map[sector],speed_reduction)/speed_reduction);
	if (stop_at_goal) 
	{
		float d = sqrt(sqr(xGoal-rs.x)+sqr(yGoal-rs.y));
		v = min(v, (float)sqrt(2*d*config.max_decel));
	}
	return v;
}

/***************************************
VFH
***************************************/


VFH::VFH(Robot *robot, int mapw, int maph): robot(robot)
{
	lm = new LocalMap(mapw, maph, config.local_map_size, config.cell_size);
	pm = new PolarMap(config.polar_sectors);
	xGoal = config.goal_x;
	yGoal = config.goal_y;
	image = new uchar[(mapw+4)*(maph+4)*3];
}

VFH::~VFH()
{
	delete lm;
	delete pm;
	delete[] image;
}




void VFH::updateMap()
{
	for(int i = 0; i < robot->nsensors; i++)
	if (robot->readings[i].updated && robot->readings[i].is_working)
	{
		/// detected point
		float range = robot->readings[i].value;
		float angle = (robot->readings[i].theta + robot->readings[i].angle)*2*M_PI/360.;
		float cos_angle = cos(angle), sin_angle = sin(angle);
		
		/// update obstacle
		bool is_obstacle = (range < robot->maxSensorRange && range != robot->sonar_outrange);
		if (!is_obstacle) range = robot->minSensorRange;
		float obstx = robot->readings[i].x + range*cos_angle; int ox = (int)(obstx/lm->cellSize);
		float obsty = robot->readings[i].y + range*sin_angle; int oy = (int)(obsty/lm->cellSize);
		if (is_obstacle)
		{
			if (ox >= 0 && ox < lm->w && oy >= 0 && oy < lm->h)
			if (lm->map[ox][oy].obst < config.max_probes)
			{
				lm->map[ox][oy].probed++;
				lm->map[ox][oy].obst++;
			}
		}
		
	INFO("Sensor %f: (%f,%f,%f) %f ; obstacle %d at (%d,%d)",
		robot->readings[i].angle, robot->readings[i].x, robot->readings[i].y, robot->readings[i].theta,
		robot->readings[i].value, (is_obstacle ? 1: 0), ox, oy);

		/// update non obstacle
		if (config.vote_noobst && robot->readings[i].value >= robot->minSensorRange)
		{
			double stepx, stepy; int steps;
			if (abs(cos_angle) > abs(sin_angle))
				{ stepx = sign(cos_angle); stepy = sin_angle/abs(cos_angle); steps = (int)(abs(obstx-robot->readings[i].x)/lm->cellSize); } else
				{ stepy = sign(sin_angle); stepx = cos_angle/abs(sin_angle); steps = (int)(abs(obsty-robot->readings[i].y)/lm->cellSize); }
				
			double x = robot->readings[i].x/lm->cellSize+3*stepx, y = robot->readings[i].y/lm->cellSize+3*stepy;
			for(int i = 3; i < steps - 3; i++, x += stepx, y += stepy)
			{
				int ix = (int)x, iy = (int)y;
				if (ix >= 0 && ix < lm->w && iy >= 0 && iy < lm->h)
				if (lm->map[ix][iy].obst > 0)
				{
					lm->map[ix][iy].probed--; 
					lm->map[ix][iy].obst--; 
				}
			}
		}
	}
}


void VFH::getDirSpeed(float &theta, float &vel, float goal_x, float goal_y, bool stop_at_goal)
{
	robot->refreshState();
	histo_thres = config.histo_thres;
	speed_reduction = config.speed_reduction;
	if (stop_at_goal) 
	{ 
		float goal_dist = sqrt(sqr(robot->state.x - goal_x) + sqr(robot->state.y - goal_y));
		if (goal_dist < 60) 
		{
			histo_thres = histo_thres+(1-goal_dist/60)*400*histo_thres;
			speed_reduction = speed_reduction+(1-goal_dist/60)*10*speed_reduction;
INFO("increased histo_thres from %f to %f", config.histo_thres, histo_thres);
		}
	}
	pm->compute(*lm, robot->state);
	theta = pm->getDirection(robot->state, goal_x, goal_y, histo_thres);
	vel = pm->getSpeed(robot->state, goal_x, goal_y, stop_at_goal, speed_reduction);
}


void VFH::saveMaps(char *filename, GlobalMap &gm, std::list<Vertex *> *list_inter, float direction)
{
	int thick = 2;
	int imgw = lm->w+2*thick, imgh = lm->h+2*thick;
	if (lm->w != gm.w || lm->h != gm.h) ERROR("local and global maps of different sizes");
	
	int idx_robot = (int)(robot->state.theta*pm->size/360);
	int idx_dir = (int)(direction*pm->size/360);
//	std::cout << "idx robot dir " << idx_robot << ", " << idx_dir << std::endl;
	
	// local and global maps
//	int num = 0;
	for(int y = 0; y < lm->h; y++) for(int x = 0; x < lm->w; x++) 
	{
		int pix = (((lm->h-1-y)+thick)*imgw+(x+thick))*3;
		image[pix+0] = 255; // default = white
		image[pix+1] = image[pix+2] = 255-(lm->map[x][y].obst*255/config.max_probes); // local map = red
		if (gm.map[x][y] == 1 && lm->map[x][y].obst == 0) { image[pix+0] = 0; image[pix+1] = 255; image[pix+2] = 0; } // global map = green
/*

		image[pix+0] = (num%255);
		image[pix+1] = (num%255);
		image[pix+2] = (num%255);
		num++;
*/

	}
	// robot position
	int pix = ((imgh-1-((int)(robot->state.y/lm->cellSize)+thick))*imgw+((int)(robot->state.x/lm->cellSize))+thick)*3;
	image[pix+0] = 0; image[pix+1] = 0; image[pix+2] = 255; 
	// planned itinerary
	for(std::list<Vertex *>::iterator it = list_inter->begin(); it != list_inter->end(); it++)
	{
		int pix = ((imgh-1-((int)((*it)->y)+thick))*imgw+((int)((*it)->x))+thick)*3;
		image[pix+0] = 0; image[pix+1] = 0; image[pix+2] = 0; 
	}
	
	// polar map plus directions
	int hL = imgh/2;
	int hl = pm->size / 8;
	float idx, val;
	int iidx, iidx1, ival;
	uchar col[3];
	
#define SET_COLOR \
		iidx = (int)idx; iidx1 = iidx+1; if (iidx1 >= pm->size) iidx1 = 0; \
		val = (1-(idx-iidx))*pm->map[iidx] + (idx-iidx)*pm->map[iidx1]; \
		ival = 255-min(255, (int)(val*255/(2*histo_thres))); \
		if (iidx == idx_dir && iidx == idx_robot) { col[0]=0; col[1]=0; col[2]=128; } else /* control direction (blue)  */ \
		if (iidx == idx_dir /*&& (idx-iidx) <= 0.5*/) { col[0]=0; col[1]=0; col[2]=255; } else /* control direction (blue)  */ \
		if (iidx == idx_robot /*&& (idx-iidx) <= 0.5*/) { col[0]=0; col[1]=0; col[2]=0; } else /* current direction of the robot (black) */ \
		if (val <= histo_thres)  { col[0]=2*ival; col[1]=255; col[2]=2*ival; } else /* valley (green)  */ \
			 { col[0]=255; col[1]=ival; col[2]=ival; } /* obstacle (red) */ \
//		{ col[0]=64; col[1]=128; col[2]=64; }

	float scale = (hL)/(float)hl;
	for(int y = hL; y < imgh; y++)
	{
		idx = (y-hL)/scale; 
		SET_COLOR
//		std::cout << "iidx " << iidx << ", iidx1 " << iidx1 << std::endl;
		for(int x = imgw-thick; x < imgw; x++) { pix=((imgh-1-y)*imgw+x)*3; image[pix+0] = col[0]; image[pix+1] = col[1]; image[pix+2] = col[2]; }
	}
	scale = (imgw-thick)/(float)(2*hl);
	for(int x = imgw-thick-1; x >= 0; x--)
	{
		idx = (imgw-thick-1-x)/scale+(hl);
		SET_COLOR
//		std::cout << "iidx " << iidx << ", iidx1 " << iidx1 << std::endl;
		for(int y = imgh-thick; y < imgh; y++) { pix=((imgh-1-y)*imgw+x)*3; image[pix+0] = col[0]; image[pix+1] = col[1]; image[pix+2] = col[2]; }
	}	
	scale = (imgh-thick)/(float)(2*hl);
	for(int y = 0; y < imgh-thick; y++)
	{
		idx = (imgh-thick-y)/scale+(3*hl);
 		SET_COLOR
//		std::cout << "iidx " << iidx << ", iidx1 " << iidx1 << std::endl;
		for(int x = 0; x < thick; x++) { pix=((imgh-1-y)*imgw+x)*3; image[pix+0] = col[0]; image[pix+1] = col[1]; image[pix+2] = col[2]; }
	}	
	scale = (imgw-thick)/(float)(2*hl);
	for(int x = thick; x < imgw; x++)
	{
		idx = (x-thick)/scale+(5*hl);
 		SET_COLOR
//		std::cout << "iidx " << iidx << ", iidx1 " << iidx1 << std::endl;
		for(int y = 0; y < thick; y++) { pix=((imgh-1-y)*imgw+x)*3; image[pix+0] = col[0]; image[pix+1] = col[1]; image[pix+2] = col[2]; }
	}	
	scale = (hL-1)/(float)(hl-1);
	for(int y = thick; y <= hL; y++)
	{
		idx = (y-thick)/scale+7*hl;
 		SET_COLOR
//		std::cout << "y " << y << ", iidx " << iidx << ", iidx1 " << iidx1 << std::endl;
		for(int x = imgw-thick; x < imgw; x++) { pix=((imgh-1-y)*imgw+x)*3; image[pix+0] = col[0]; image[pix+1] = col[1]; image[pix+2] = col[2]; }
	}
//	std::cout << "hL " << hL << ", imgh " << imgh << std::endl;

	// active window (yellow)
	int activex = (int)(robot->state.x/lm->cellSize) - config.active_window_size/2;
	int activey = (int)(robot->state.y/lm->cellSize) - config.active_window_size/2;
//	std::cout << "active xy " << activex << "," << activey << std::endl;
	for(int x = 0; x < config.active_window_size; x++)
	{
		if (x <= -activex || x+activex>=lm->w) continue;
		for(int y = 0; y < config.active_window_size; y++)
		{
			if (y <= -activey && y+activey>=lm->h) continue;
			int pix = ((imgh-1-(activey+y+2))*imgw+(activex+x+2))*3;
			if (image[pix+0] == 255 && image[pix+1] == 255 && image[pix+2] == 255) image[pix+2] = 0;
		}
	}
	
	// ok we can write the image !
	FILE *file = fopen(filename, "wb"); ASSERT(file);
	fprintf(file, "P%c\n%d %d\n%d\n", '6', imgw, imgh, 255);
	int res = fwrite(image, 1, imgw*imgh*3, file);
	ASSERTC(res,==,imgw*imgh*3);
	fclose(file);	
}




